from __future__ import annotations
import asyncio
import aiohttp
from datetime import datetime

class Forecast:
    def __init__(self, totals: dict, grid_x: int, grid_y: int, cwa: str, city: str, state: str):
        self.totals = totals
        self.simplified_dict = {key: '...' for key in totals.keys()}
        self.grid_x, self.grid_y = grid_x, grid_y
        self.city, self.state = city, state
        self.cwa = cwa

    def __repr__(self):
        return f"Forecast(totals={self.simplified_dict}, grid_x={self.grid_x}, grid_y={self.grid_y}," \
               f" cwa={self.cwa}, city={self.city}, state={self.state})"


def forecast(latitude: float, longitude: float) -> list[Forecast]:
    """Main function for retrieving forecast for a longitude and latitude"""
    loop = asyncio.get_event_loop()
    *information, request = loop.run_until_complete(_get_gridpoint(latitude, longitude))
    res = loop.run_until_complete(_get_forecast(request))
    return Forecast(totals=res, grid_x=information[0], grid_y=information[1],
                    cwa=information[2], city=information[3], state=information[4])


async def _get_gridpoint(latitude: float, longitude: float) -> tuple[int, int, str, str, str, str]:
    async with aiohttp.ClientSession() as session:
        async with session.get(f"https://api.weather.gov/points/{latitude},{longitude}") as response:
            response = await response.json()
            response = response['properties']
            return response['gridX'], response['gridY'], response['gridId'], \
                   response['relativeLocation']['properties']['city'], \
                   response['relativeLocation']['properties']['state'], \
                   response['forecastGridData']


async def _get_forecast(url: str) -> Forecast:
    async with aiohttp.ClientSession() as session:
        async with session.get(url) as response:
            response = await response.json()
            response = response['properties']
            totals = {}
            for key, val in zip(list(response.keys())[8:], list(response.values())[8:]):
                if isinstance(val, dict):
                    totals[key] = {}
                    for time_dct in val['values']:
                        time = time_dct['validTime'].split('/')[0]
                        totals[key][datetime.fromisoformat(time)] = time_dct['value']

            return totals


from __future__ import annotations
import asyncio
import aiohttp
from itertools import chain, zip_longest
import inspect
from datetime import datetime


class Alert:
    """Class to show a type of alert."""
    __slots__ = ('coordinates', 'areas_affected', 'affected_zones', 'start_time', 'end_time', 'cwa_issued',
                 'status', 'message_type', 'category', 'severity', 'certainty', 'urgency', 'event',
                 'description', 'instructions')

    def __init__(self, coordinates: list, area_desc: str, zones: list, starts: str, ends: str,
                 cwa: str, status: str, message_type: str, category: str, severity: str, certainty: str,
                 urgency: str, event: str, description: str, instructions: str):
        self.coordinates = list(chain.from_iterable(coordinates)) if coordinates else None
        self.areas_affected = area_desc
        self.affected_zones = zones
        self.start_time = datetime.fromisoformat(starts) if starts else None
        self.end_time = datetime.fromisoformat(ends) if ends else None
        self.cwa_issued = cwa
        self.status = status
        self.message_type = message_type
        self.category = category
        self.severity = severity
        self.certainty = certainty
        self.urgency = urgency
        self.event = event
        self.description = description
        self.instructions = instructions


def alerts(active: bool = True, point: tuple = None, events: tuple = None, start: str = None, end: str = None,
           status: str = None, codes: tuple = None, message_type: str = None, region_type: str = None,
           region: str = None, area: str = None, zone: str = None, urgency: str = None, severity: str = None,
           certainty: str = None, limit: int = None) -> None:
    """Retrieve alerts from the api.weather.gov API with numerous parameters to filter through current alerts."""
    if any((True for param, val in locals().items() if (isinstance(val, tuple) if
    param not in ('point', 'events', 'codes') else (isinstance(val[0], tuple) if val else None)))):
        loop = asyncio.get_event_loop()
        res = loop.run_until_complete(_multiple_alerts(active, point, events, start, end, status, codes, message_type,
                                                       region_type, region, area, zone, urgency, severity,
                                                       certainty, limit))
        return res
    else:
        loop = asyncio.get_event_loop()
        res = loop.run_until_complete(_get_alert(active, point, events, start, end, status, codes, message_type,
                                                 region_type, region, area, zone, urgency, severity, certainty, limit))
        return res


async def _multiple_alerts(active: bool, point: tuple, events: tuple, start: str, end: str,
                           status: str, codes: tuple, message_type: str, region_type: str,
                           region: str, area: str, zone: str, urgency: str, severity: str,
                           certainty: str, limit: int):
    """Recieves multiple alerts"""

    to_list = dict(locals())
    for var, value in to_list.items():
        if not isinstance(value, list) and not isinstance(value, tuple):
            to_list[var] = [value]

    funcs_to_run = [_get_alert(act, poi, even, st, en, stat, code, msg_type,
                               reg_type, reg, ar, zon, urgen, sever, certain, lim) for act, poi, even, st, en,
                                                                                       stat, code, msg_type, reg_type, reg, ar, zon, urgen, sever, certain, lim
                    in zip_longest(
            to_list['active'], to_list['point'], to_list['events'], to_list['start'], to_list['end'],
            to_list['status'], to_list['codes'], to_list['message_type'], to_list['region_type'],
            to_list['region'], to_list['area'], to_list['zone'], to_list['urgency'],
            to_list['severity'], to_list['certainty'], to_list['limit']
        )]
    results = await asyncio.gather(
        *funcs_to_run
    )
    return results


async def _get_alert(active: bool, point: tuple, event: tuple, start: str, end: str,
                     status: str, code: tuple, message_type: str, region_type: str,
                     region: str, area: str, zone: str, urgency: str, severity: str,
                     certainty: str, limit: int):
    """Retrieves the actual alert from the API itself as a helper function that is aysnchronous for speed-ups."""
    return_result = []
    pars = locals()
    url = "https://api.weather.gov/alerts"
    for idx, arg in enumerate(inspect.getargs(_get_alert.__code__).args):
        if str(pars.get(arg)) not in ('None', '0'):
            url += (f'?{arg}=' if not idx else f'&{arg}=')
        results = {True: 'true', False: 'false', None: 'null'}
        if str((res := pars.get(arg))) not in ('None', '0'):
            if isinstance(res, tuple):
                url += r'%2C'.join(tuple(map(str, pars.get(arg))))
            elif res in results.keys():
                url += results[pars.get(arg)]
            else:
                url += pars.get(arg)

    async with aiohttp.ClientSession() as session:
        async with session.get(url) as response:
            response = await response.json()
            if 'detail' in response.keys():
                raise TypeError(response['detail'])
            for alert in response['features']:
                if alert['geometry']:
                    geom = alert['geometry']['coordinates']
                else:
                    geom = None
                temp_resp = alert['properties']
                return_result.append(Alert(area_desc=temp_resp['areaDesc'], zones=temp_resp['affectedZones'],
                                           starts=temp_resp['onset'], ends=temp_resp['expires'],
                                           status=temp_resp['status'], message_type=temp_resp['messageType'],
                                           category=temp_resp['category'], severity=temp_resp['severity'],
                                           certainty=temp_resp['certainty'], description=temp_resp['description'],
                                           urgency=temp_resp['urgency'], event=temp_resp['event'],
                                           cwa=temp_resp['senderName'], instructions=temp_resp['instruction'],
                                           coordinates=geom))

    return return_result


from enum import Enum


class Weather(Enum):
    snow = 1
    max_temp = 2
    min_temp = 3
    precip = 4
    max_gust = 5
    snow_depth = 6
    min_rh = 7
    max_rh = 8
    max_dewpoint = 9
    min_dewpoint = 10
    min_heat_index = 11
    max_heat_index = 12
    avg_heat_index = 13



import asyncio
import aiohttp
from datetime import datetime
from weather_types import Weather
from typing import Union


states = """
Alabama	AL
Alaska	AK
Arizona	AZ
Arkansas	AR
California	CA
Colorado	CO
Connecticut	CT
Delaware	DE
Florida	FL
Georgia	GA
Hawaii	HI
Idaho	ID
Illinois	IL
Indiana	IN
Iowa	IA
Kansas	KS
Kentucky	KY
Louisiana	LA
Maine	ME
Maryland	MD
Massachusetts	MA
Michigan	MI
Minnesota	MN
Mississippi	MS
Missouri	MO
Montana	MT
Nebraska	NE
Nevada	NV
New Hampshire	NH
New Jersey	NJ
New Mexico	NM
New York	NY
North Carolina	NC
North Dakota	ND
Ohio	OH
Oklahoma	OK
Oregon	OR
Pennsylvania	PA
Rhode Island	RI
South Carolina	SC
South Dakota	SD
Tennessee	TN
Texas	TX
Utah	UT
Vermont	VT
Virginia	VA
Washington	WA
West Virginia	WV
Wisconsin	WI
Wyoming	WY
"""
abbreviations = {l.strip('\n').lower(): k.strip('\n') for l, k in
                     [(line.split('\t')[0],line.split('\t')[-1].strip('\n')) for line in states]}
abbreviations.update({'district of columbia': 'DC'})

abbreviations_change = {state: abbr for state, abbr in zip(abbreviations.values(), abbreviations.keys())}


class HistoricalWeatherData:
    def __init__(self, airport, state):
        self.airport = airport.lstrip('K')
        self.state = state if len(state) == 2 else abbreviations[state.lower()]

    async def _calculate(self, date: tuple, enum_type: Weather) -> dict:
        """Calculate the requested historical data"""
        if isinstance(date, int) or isinstance(date, str):
            raise ValueError("Tuple expected for date parameter.\n\n"
                             f"Did you mean: ...generate(({date},), {enum_type})")
        mapping_key = {Weather.max_temp: 'max_tmpf', Weather.min_temp: 'min_tempf',
                       Weather.precip: 'precip', Weather.max_gust: 'max_gust', Weather.snow: 'snow',
                       Weather.snow_depth: 'snowd', Weather.min_rh: 'min_rh', Weather.max_rh: 'max_rh',
                       Weather.max_dewpoint: 'max_dwpf', Weather.min_dewpoint: 'min_dwpf',
                       Weather.max_heat_index: 'max_feel', Weather.min_heat_index: 'min_feel',
                       Weather.avg_heat_index: 'avg_feel'}
        enum_type = mapping_key[enum_type]
        async with aiohttp.ClientSession() as session:
            url = self.generate_url(date)
            async with session.get(url) as response:
                iem_request = await response.json()
                iem_request = iem_request['data']
                expr = {datetime.strptime(data['date'], '%Y-%m-%d'): data[enum_type]
                        for data in iem_request if data[enum_type]}
                return expr

    async def _get_total_snow(self, date: tuple, type: Union[tuple[Weather], Weather]):
        if isinstance(type, tuple):
            all_res = [self._calculate(date, enum_type) for enum_type in type]
        else:
            all_res = [self._calculate(date, type)]
            type = [type]

        total = await asyncio.gather(*all_res)
        total_dct = {}
        for result, type_wanted in zip(total, type):
            total_dct[type_wanted.name] = result

        return total_dct

    def generate(self, date: tuple, type: Union[tuple[Weather], Weather]):
        loop = asyncio.get_event_loop()
        data = loop.run_until_complete(self._get_total_snow(date, type))
        return data

    def generate_url(self, date: tuple) -> str:
        """Generates url for self._calculate"""
        if len(date) == 3:
            if len(str(date[1])) == 1:
                date = list(date)
                date[1] = '0' + str(date[1])
            url = f"https://mesonet.agron.iastate.edu/api/1/" \
                  f"daily.json?date={date[0]}-{date[1]}-{date[2]}&" \
                  f"station={self.airport}&network={self.state}_ASOS"
        elif len(date) == 2:
            if len(str(date[1])) == 1:
                date = list(date)
                date[1] = '0' + str(date[1])
            url = f"https://mesonet.agron.iastate.edu/api/1/daily.json?" \
                  f"station={self.airport}&network={self.state}_ASOS&year={date[0]}&month={date[1]}"
        else:
            url = f'https://mesonet.agron.iastate.edu/api/1/daily.json?' \
                  f'station={self.airport}&network={self.state}_ASOS&year={date[0]}'

        return url



from __future__ import annotations
import asyncio
import aiohttp
from inflection import underscore
from typing import Union
from datetime import datetime


def latest_surface_observations(latitude: int = 0, longitude: int = 0, station: str = '',
                                imperial: bool = False, start: datetime = '',
                                end: datetime = '', limit: int = None) -> dict:
    """Get the latest surface observations for the closest station."""
    loop = asyncio.get_event_loop()
    if isinstance(start, datetime):
        start = start.isoformat() + '%2B00:00'
    if isinstance (end, datetime):
        end = end.isoformat() + '%2B00:00'

    if isinstance(latitude, tuple) or isinstance(longitude, tuple) or isinstance(station, tuple):
        result = loop.run_until_complete(multiple_surface_obs(latitude, longitude, station,
                                                              imperial, start, end, limit))
    else:
        if not station and not latitude and not longitude:
            raise TypeError("This function expected either a station name, or a latitude and longtitude provided.")
        elif station:
            station = f'https://api.weather.gov/stations/{station}'
        elif not station:
            station = loop.run_until_complete(_get_closest_station(latitude, longitude))

        result = loop.run_until_complete(_calculate_lso(station, imperial, start, end, limit))
    return result


async def multiple_surface_obs(latitude: Union[int, tuple], longitude: Union[int, tuple],
                               station: Union[str, tuple], imperial: bool, start: str, end: str, limit: int):
    if isinstance(station, tuple):
        station = [f'https://api.weather.gov/stations/{station_name}' for station_name in station]
    if isinstance(latitude, tuple) and isinstance(longitude, tuple):
        if len(latitude) != len(longitude):
            raise ValueError("The latitude and longitude tuples need to be of the same length.")

        lat_long = zip(latitude, longitude)
        station_funcs = [_get_closest_station(lat, long) for lat, long in lat_long]
        station = await asyncio.gather(
            *station_funcs
        )

    result_funcs = [_calculate_lso(stat, imperial, start, end, limit) for stat in station]
    results = await asyncio.gather(
        *result_funcs
    )

    result = results[0]
    [result.update(other_result) for other_result in results[1:]]
    return result

async def _get_closest_station(latitude: int, longitude: int) -> str:
    """Gets the closest weather station"""
    async with aiohttp.ClientSession() as session:
        async with session.get(f'https://api.weather.gov/points/{latitude},{longitude}/stations') as response:
            closest_station = await response.json()
            return closest_station['features'][0]['id']


async def _calculate_lso(station: str, imperial: bool, start: str, end: str, limit: int) -> dict:
    """Finds latest observations"""
    fahrenheit = lambda x: round(x * 1.8 + 32, 5)
    miles = lambda x: round(0.00062137 * x, 5)
    miles_per_hour = lambda x: round(0.621371 * x, 5)
    inches = lambda x: round(39.3701 * x, 5)
    millibars = lambda x: round(x / 100, 5)
    station_name = station.split('/')[-1]
    obs_dct = {station_name: [] if start and end else {}}
    url = station + (f'/observations?start={start}&end={end}' if start and end else '/observations/latest')
    if '?start' in url and '&end' in url and limit:
        url += f'&limit={limit}'
    async with aiohttp.ClientSession() as session:
        async with session.get(url) as response:
            response = await response.json(content_type=None)
            if 'status' in response.keys() and response['status'] == 404 and (start and end):
                raise ValueError(f"{station} did not have any "
                                 f"observations from {start} to {end}.")

            if not start and not end:
                for key, val in response['properties'].items():
                    if key in {'temperature', 'dewpoint', 'windDirection', 'windSpeed',
                               'windGust', 'barometricPressure', 'seaLevelPressure',
                               'visibility', 'precipitationLastHour', 'precipitationLast3Hours',
                               'precipitationLast6Hours', 'relativeHumidity', 'windChill',
                               'heatIndex'}:
                        if imperial and 'degC' in val['unitCode'] and val['value']:
                            obs_dct[station_name][underscore(key)] = fahrenheit(val['value'])
                        elif imperial and 'unit:m' in val['unitCode'] and val['value']:
                            if key == 'visibility':
                                obs_dct[station_name][underscore(key)] = miles(val['value'])
                            else:
                                obs_dct[station_name][underscore(key)] = inches(val['value'])
                        elif imperial and 'unit:km_h-1' in val['unitCode'] and val['value']:
                            obs_dct[station_name][underscore(key)] = miles_per_hour(val['value'])
                        elif imperial and 'unit:Pa' in val['unitCode'] and val['value']:
                            obs_dct[station_name][underscore(key)] = millibars(val['value'])
                        else:
                            obs_dct[station_name][underscore(key)] = val['value']
            else:
                for obs in response['features']:
                    obs_dct[station_name].append({})
                    cur_dct = obs_dct[station_name][-1]
                    cur_dct['time'] = datetime.fromisoformat(obs['properties']['timestamp'])
                    for key, val in obs['properties'].items():
                        if key in {'temperature', 'dewpoint', 'windDirection', 'windSpeed',
                                   'windGust', 'barometricPressure', 'seaLevelPressure',
                                   'visibility', 'precipitationLastHour', 'precipitationLast3Hours',
                                   'precipitationLast6Hours', 'relativeHumidity', 'windChill',
                                   'heatIndex'}:
                            if imperial and 'degC' in val['unitCode'] and val['value']:
                                cur_dct[underscore(key)] = fahrenheit(val['value'])
                            elif imperial and 'unit:m' in val['unitCode'] and val['value']:
                                if key == 'visibility':
                                    cur_dct[underscore(key)] = miles(val['value'])
                                else:
                                    cur_dct[underscore(key)] = inches(val['value'])
                            elif imperial and 'unit:km_h-1' in val['unitCode'] and val['value']:
                                cur_dct[underscore(key)] = miles_per_hour(val['value'])
                            elif imperial and 'unit:Pa' in val['unitCode'] and val['value']:
                                cur_dct[underscore(key)] = millibars(val['value'])
                            else:
                                cur_dct[underscore(key)] = val['value']

    return obs_dct


def compare(dct_to_compare: dict) -> dict:
    """Compares dictionary of existing surface observations, as intended to."""
    return {key: sorted([(dct_to_compare[station][key], station) for station in dct_to_compare.keys()],
                        key=lambda x: x[0] if x[0] else 0)
            for key in list(dct_to_compare.values())[0].keys()}


import pathlib
from setuptools import setup, find_packages

CURRENT_DIRECTORY = pathlib.Path(__file__).parent

# The text of the README file
README = (CURRENT_DIRECTORY / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="weather_usa",
    version="1.1.0",
    description="Wrapper for parts of the IEM and NWS API, built with asyncio for speed",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/Shom770/weather_usa",
    license="MIT",
    install_requires=["aiohttp", "inflection"],
    packages=find_packages()
)